/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Copyright (C) 2011-2018 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::structureModel

Description
    Class to handle the description of a mesh-unresolved structure. It is 
    derived from phaseBase, to represent the volume fraction of the structure.
    The key member for the understanding of this class is cellFields_. 
    The structure can be in fact though of as being composed by different
    regions, with properties defined only on certain cells of the mesh.
    However, as these region are meant NOT to spatially overlap, the 
    overall structure can be described by a single set of fields (Dh_, iApow_, 
    etc.) that are initialized/set/modified on a cell-by-cell basis.
    The cellFields_ member is thus a HashTable of scalarFields indexed by
    region name. These scalarFields are = 1.0 where the given
    region exists, and 0 elsewhere. What the region names are, that
    is set by the actual run-time-selected instance of the structureModel
    class.
    Currently, the only available structureModel is byZone, in which
    the regions are effectively cellZones, and the cellFields_ is
    thus a dicts of scalarFields representing where each cellZone exists.


Author 
    Stefan Radman (stefan.radman@epfl.ch / stefanradman92@gmail.com)

SourceFiles
    structureModel.C

\*---------------------------------------------------------------------------*/

#ifndef structureModel_H
#define structureModel_H

#include "fvMatrix.H"
#include "phaseBase.H"
#include "HashPtrTable.H"
#include "runTimeSelectionTables.H"
#include "zeroGradientFvPatchFields.H"
#include "pump.H"
#include "heatExchanger.H"
#include "powerModel.H"
#include "powerOffCriterionModel.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

//- Forward declarations
class fluid;

/*---------------------------------------------------------------------------*\
                           Class structureModel Declaration
\*---------------------------------------------------------------------------*/

class structureModel
:
    public phaseBase
{
protected:

    typedef HashTable
    <
        pump,
        word,
        word::hash
    >   pumpTable;

    typedef HashTable
    <
        heatExchanger,
        word,
        word::hash
    >   heatExchangerTable;

    typedef HashTable
    <
        autoPtr<powerModel>,
        word,
        word::hash
    >   powerModelTable;

    //- List of names of each group of cells into which this structure is split
    wordList regions_;

    //- Table of limited volume fraction fields indexed by region name
    //  Every field is equal to the overall alpha of structure only in those
    //  cells where the corresponding region is present, and is 0 
    //  elsewhere
    volScalarFieldTable alphaFields_;

    //- Table of cell scalarFields indexed by region name
    scalarFieldTable cellFields_;

    //- Table of cell labelLists indexed by region name
    labelListTable cellLists_;

    //- All cells in which the structure exists ()
    labelList cells_;

    //- Tortuosity of the structure in the global reference frame
    volTensorField tortuosity_;

    //- Hydraulic diameter
    volScalarField Dh_;

    //- Hydraulic diameter in the X, Y, Z directions in the local structure 
    //  frame of reference. Used exclusively for the computation of the local
    //  vector Reynolds number.
    volVectorField lDh_;

    //- Table of momentum sources (e.g. pumps)
    pumpTable pumps_;

    //- Table of heat exchangers
    heatExchangerTable heatExchangers_;

    //- Table of power models indexed by powerModel name
    powerModelTable powerModels_;

    //- Heat flux from active structures only. Positive if from structure to
    //  fluid
    volScalarField heatFlux_;

    //- Surface temperature of power regions
    volScalarField Tact_;

    //- Interfacial area of the powerModel region
    volScalarField iAact_;

    //- Heat exchanger surface temperature
    autoPtr<volScalarField> THXPtr_;

    //- Heat exchanger interfacial area (no difference between inner and outer,
    //  don't ask too much pls thnx. Maybe in the future)
    autoPtr<volScalarField> iAHXPtr_;

    //- Volume fraction of the passive structure (might be different than
    //  global structure volumefraction)
    volScalarField alphapas_;

    //- Surface temperature of the passive region
    volScalarField Tpas_;

    //- Interfacial area of the passive region
    volScalarField iApas_;

    //- Volumetric heat capacity of the passive region times volume
    //  fraction of passive region
    volScalarField alphaRhoCppas_;

    //- Max between Tact_ and Tpas_;
    volScalarField Twall_;

    //- Rotation matrix to move from global to local reference frame
    volTensorField Rg2l_;

    //- Rotation matrix to move from local to global reference frame
    volTensorField Rl2g_;

    //- Total momentum source
    autoPtr<volVectorField> momentumSourcePtr_;

    //- Power off criterion
    autoPtr<powerOffCriterionModel> powerOffCriterionModelPtr_;

public:

    TypeName("structureModel");

    //- Runtime selection table declaration

        declareRunTimeSelectionTable
        (
            autoPtr,
            structureModel,
            structureModels,
            (
                const dictionary& dict,
                const fvMesh& mesh
            ),
            (dict, mesh)
        );

    //- Constructors

        structureModel
        (
            const dictionary& dict,
            const fvMesh& mesh
        );

        //- Destructor
        virtual ~structureModel(){}

    //- Static member functions

        //- Selector

        static autoPtr<structureModel> New
        (
            const dictionary& dict,
            const fvMesh& mesh
        );

    //- Member Functions

        template<class Type>
        void setRegionField
        (
            const dictionary& dicts,
            GeometricField<Type, fvPatchField, volMesh>& field,
            word key
        ) const;

        //- Adjust pumps and return momentum source
        const volVectorField& momentumSource();

        //- By default heatExchangers operate on cellZones, so the way they are
        //  constructed is handled by a function here rather than in byZone
        void constructHeatExchangers();

        //- In two-phase solvers :
        //  HT = H1*frac1*T1 + H2*frac2*T2
        //  H =  H1*frac1 + H2*frac2
        //- In a mono-phase solver :
        //  HT = H*T
        //  H = H
        virtual void correct
        (
            const volScalarField& HT,
            const volScalarField& H
        );

        tmp<volScalarField> explicitHeatSource
        (
            const fluid& fluid,
            const volScalarField& H
        ) const;

        tmp<fvScalarMatrix> linearizedSemiImplicitHeatSource
        (
            fluid& fluid,
            const volScalarField& H
        ) const;

        //- Function to access any member field of a powerModel, needed e.g.
        //  to get temperatures from the fuelModel

        //- Turn off all powerDensity fields of all the powerModels if the
        //  powerOffCriterionModel criterion is satisfied
        void checkPowerOff();

        //- Const access

            const wordList& regions() const
            {
                return regions_;
            }

            const volScalarFieldTable& alphaFields() const
            {
                return alphaFields_;
            }

            const scalarFieldTable& cellFields() const
            {
                return cellFields_;
            }

            const labelListTable& cellLists() const
            {
                return cellLists_;
            }

            const labelList& cells() const
            {
                return cells_;
            }

            const volTensorField& tortuosity() const
            {
                return tortuosity_;
            }

            const volScalarField& Dh() const
            {
                return Dh_;
            }

            const volVectorField& lDh() const
            {
                return lDh_;
            }

            const volScalarField& Tact() const
            {
                return Tact_;
            }

            const volScalarField& iAact() const
            {
                return iAact_;
            }

            const volScalarField& iApas() const
            {
                return iApas_;
            }

            const volScalarField& Tpas() const
            {
                return Tpas_;
            }

            const volScalarField& Twall() const
            {
                return Twall_;
            }

            const volTensorField& Rg2l() const
            {
                return Rg2l_;
            }

            const volTensorField& Rl2g() const
            {
                return Rl2g_;
            }

            const autoPtr<volVectorField>& momentumSourcePtr() const
            {
                return momentumSourcePtr_;
            }
};

//- Definition of templated static function in header, it's best practice
//  according to C++ 11 guidelines
template<class Type>
void structureModel::setRegionField 
(
    const dictionary& dicts,
    GeometricField<Type, fvPatchField, volMesh>& field,
    word key
) const
{
    typedef GeometricField<Type, fvPatchField, volMesh> fieldType;

    IOobject fieldHeader
    (
        field.name(),
        mesh_.time().timeName(),
        mesh_,
        IOobject::NO_READ
    );

    //- field is a reference to a field that already exists. If it was
    //  created with a READ_IF_PRESENT or MUST_READ option, it is not a bad
    //  idea to spatially limit the field only over the cells overwhich it
    //  should be defined, just in case
    if 
    (
        (
            field.readOpt() == IOobject::MUST_READ
        or  field.readOpt() == IOobject::READ_IF_PRESENT
        )
    and fieldHeader.typeHeaderOk<fieldType>(true)
    )
    {
        scalarField cellField(mesh_.cells().size(), 0);
        forAll(dicts.toc(), i)
        {
            word region(dicts.toc()[i]);
            const labelList& cells(cellLists_[region]);
            forAll(cells, j)
            {
                cellField[cells[j]] = 1.0;
            }
        }
        field.primitiveFieldRef() *= cellField;
    }
    //- Otherwise, read the value the field should have on a 
    //  region-by-region basis
    else
    {
        forAll(dicts.toc(), i)
        {
            word region(dicts.toc()[i]);
            const labelList& cells(cellLists_[region]);
            Type value(dicts.subDict(region).get<Type>(key));
            forAll(cells, j)
            {
                field[cells[j]] = value;
            }
        }
    }
    field.correctBoundaryConditions();
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
